/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.io.IOException;
import java.io.EOFException;
import java.io.File;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Vector;
import java.nio.file.Path;
import java.nio.file.Files;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.primitives.*;
import com.dickimawbooks.texparserlib.generic.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;

public class AbstractParser extends LaTeXParserListener
  implements Writeable
{
   public AbstractParser(TeXApp texApp)
     throws IOException
   {
      super(null, true);
      this.texApp = texApp;

      setWriteable(this);

   }

   protected void addPredefined()
   {
      super.addPredefined();

      parser.putControlSequence(new ArticleAbstractDec());
      parser.putControlSequence(new AtSecondOfTwo("hyperlink"));
      parser.putControlSequence(new AtSecondOfTwo("hypertarget"));
      parser.putControlSequence(new EndInput("tableofcontents"));

      parser.putActiveChar(new NoExpandActiveChar('~'));
   }

   public TeXApp getTeXApp()
   {
      return texApp;
   }

   // Ignore unknown commands
   public ControlSequence createUndefinedCs(String name)
   {
      return new Undefined(name)
      {
         public void process(TeXParser parser) throws IOException
         {
         }
      };
   }

   public TeXParser parse(File texFile)
     throws IOException
   {
      TeXParser parser = new TeXParser(this);

      parser.parse(texFile);

      return parser;
   }

   public Writeable getWriteable()
   {
      return this;
   }

   public void write(String text)
     throws IOException
   {
   }

   public void writeln(String text)
     throws IOException
   {
   }

   public void write(char c)
     throws IOException
   {
   }

   public void writeCodePoint(int codePoint)
     throws IOException
   {
   }

   public void overwithdelims(TeXObject firstDelim,
     TeXObject secondDelim, TeXObject before, TeXObject after)
    throws IOException
   {
   }

   public void abovewithdelims(TeXObject firstDelim,
     TeXObject secondDelim, TeXDimension thickness, TeXObject before, 
     TeXObject after)
    throws IOException
   {
   }

   public void skipping(Ignoreable ignoreable)
      throws IOException
   {
   }

   public void substituting(String original, String replacement)
   {
      texApp.substituting(parser, original, replacement);
   }

   public void href(String url, TeXObject text)
      throws IOException
   {
   }

   public void includegraphics( 
     KeyValList options, String imgName)
     throws IOException
   {
   }

   public void subscript(TeXObject arg)
     throws IOException
   {
   }

   public void superscript(TeXObject arg)
     throws IOException
   {
   }

   public void endParse(File file)
      throws IOException
   {
      this.file = null;
   }

   public void beginParse(File file)
      throws IOException
   {
      this.file = file;
   }

   public File getFile()
   {
      return file;
   }

   public void addToAbstract(TeXObject obj)
   {
      abstractContents.add(obj);
   }

   public TeXObjectList getAbstractContents()
   {
      return abstractContents;
   }

   public String getAbstract()
      throws IOException
   {
      return abstractContents.toString(parser).trim();
   }

   public String getTitle()
    throws IOException
   {
      ControlSequence cs = getControlSequence("@title");

      TeXObjectList expanded = null;

      if (cs instanceof Expandable)
      {
         expanded = ((Expandable)cs).expandfully(getParser());
      }

      return (expanded == null ? "" : expanded.toString(getParser()));
   }

   public String getBibAuthors()
      throws IOException
   {
      LaTeXFile lfile = getDocumentClass();

      if (lfile == null || !(lfile instanceof JmlrCls))
      {
         return "";
      }

      Vector<Group> authors = ((JmlrCls)lfile).getAuthors();

      if (authors == null)
      {
         ControlSequence cs = getControlSequence("@author");

         cs.process(getParser());

         authors = ((JmlrCls)lfile).getAuthors();

         if (authors == null && cs instanceof Expandable)
         {
            TeXObjectList expanded = ((Expandable)cs).expandfully(getParser());

            if (expanded != null)
            {
               Group name = createGroup();

               for (int i = 0, n = expanded.size(); i < n; i++)
               {
                  TeXObject obj = expanded.get(i);

                  if ((obj instanceof CharObject
                       && ((CharObject)obj).getCharCode() == (int)',')
                   || (obj instanceof ControlSequence
                       && ((ControlSequence)obj).getName().toLowerCase().equals("and")))
                  {
                     if (authors == null)
                     {
                        authors = new Vector<Group>();
                     }

                     authors.add(name);
                     name = createGroup();

                     if (i+5 < n)
                     {
                        TeXObject next1 = expanded.get(i+1);
                        TeXObject next2 = expanded.get(i+2);
                        TeXObject next3 = expanded.get(i+3);
                        TeXObject next4 = expanded.get(i+4);
                        TeXObject next5 = expanded.get(i+5);

                        if (next1 instanceof WhiteSpace
                         && next2 instanceof CharObject
                         && ((CharObject)next2).getCharCode() == (int)'a'
                         && next3 instanceof CharObject
                         && ((CharObject)next3).getCharCode() == (int)'n'
                         && next4 instanceof CharObject
                         && ((CharObject)next4).getCharCode() == (int)'d'
                         && next5 instanceof WhiteSpace)
                        {
                           i += 5;
                        }
                     }
                  }
                  else if (obj instanceof WhiteSpace)
                  {
                     if (name.isEmpty()) continue;

                     if (i+4 < n)
                     {
                        TeXObject next1 = expanded.get(i+1);
                        TeXObject next2 = expanded.get(i+2);
                        TeXObject next3 = expanded.get(i+3);
                        TeXObject next4 = expanded.get(i+4);

                        if (next1 instanceof CharObject
                         && ((CharObject)next1).getCharCode() == (int)'a'
                         && next2 instanceof CharObject
                         && ((CharObject)next2).getCharCode() == (int)'n'
                         && next3 instanceof CharObject
                         && ((CharObject)next3).getCharCode() == (int)'d'
                         && next4 instanceof WhiteSpace)
                        {
                           if (authors == null)
                           {
                              authors = new Vector<Group>();
                           }

                           authors.add(name);
                           name = createGroup();
                           i += 4;
                           continue;
                        }
                     }

                     name.add(obj);
                  }
                  else
                  {
                     name.add(obj);
                  }
               }

               if (authors == null)
               {
                  authors = new Vector<Group>();
               }

               authors.add(name);
               name = createGroup();
            }
         }
      }

      if (authors == null) return "";

      StringBuffer buff = new StringBuffer();

      for (int i = 0; i < authors.size(); i++)
      {
         Group author = authors.get(i);

         if (i > 0)
         {
            buff.append(" and ");
         }

         TeXObjectList surname = new TeXObjectList();

         int n = author.size()-1;

         for (; n >= 0; n--)
         {
            TeXObject obj = author.get(n);

            if (obj instanceof WhiteSpace)
            {
               break;
            }

            surname.push(obj);
         }

         if (surname.size() == 1)
         {
            TeXObject obj = surname.get(0);

            if (obj instanceof Group)
            {
               obj = ((Group)obj).toList();
            }

            buff.append(obj.toString(getParser()));
         }
         else
         {
            buff.append(surname.toString(getParser()));
         }

         buff.append(", ");

         for (int j = 0; j < n; j++)
         {
            buff.append(author.get(j).toString(getParser()));
         }
      }

      return buff.toString();
   }

   private TeXApp texApp;

   private File file;

   private TeXObjectList abstractContents = new TeXObjectList();
}

class ArticleAbstractDec extends AbstractDec
{
   public ArticleAbstractDec()
   {
      super();
   }

   public ArticleAbstractDec(String name)
   {
      super(name);
   }

   public Object clone()
   {
      return new ArticleAbstractDec(getName());
   }

   public void process(TeXParser parser) throws IOException
   {
      process(parser, parser);
   }

   public void process(TeXParser parser, TeXObjectList stack) throws IOException
   {
      TeXObject object;
      AbstractParser listener = ((AbstractParser)parser.getListener());

      while ((object = stack.pop()) != null)
      {
         if (object instanceof TeXCsRef)
         {
            object = listener.getControlSequence(((TeXCsRef)object).getName());
         }

         if (object instanceof Expandable)
         {
            TeXObjectList expanded;

            if (stack == parser)
            {
               expanded = ((Expandable)object).expandfully(parser);
            }
            else
            {
               expanded = ((Expandable)object).expandfully(parser, stack);
            }

            if (expanded != null && !expanded.isEmpty())
            {
               object = expanded.pop();

               if (!expanded.isEmpty())
               {
                  stack.addAll(0, expanded);
               }
            }
         }

         if (object instanceof ControlSequence)
         {
            if (((ControlSequence)object).getName().equals("end"))
            {
               TeXObject arg =
                (stack == parser ? parser.popNextArg() : stack.popArg(parser));

               if (arg != null
                && arg.toString(parser).equals("abstract"))
               {
                  throw new EOFException();
               }

               listener.addToAbstract(object);

               Group grp = listener.createGroup();

               if (arg instanceof TeXObjectList)
               {
                  grp.addAll(0, (TeXObjectList)arg);
               }
               else
               {
                  grp.add(arg);
               }

               listener.addToAbstract(grp);
               continue;
            }
         }

         listener.addToAbstract(object);
      }
   }

   public void end(TeXParser parser)
    throws IOException
   {
      throw new EOFException();
   }

}

class NoExpandActiveChar extends ActiveChar
{
   public NoExpandActiveChar(int charCode)
   {
      this.charCode = charCode;
   }

   public NoExpandActiveChar(char c)
   {
      this.charCode = (int)c;
   }

   public Object clone()
   {
      return new NoExpandActiveChar(getCharCode());
   }

   public int getCharCode()
   {
      return charCode;
   }

   public TeXObjectList expandonce(TeXParser parser) throws IOException
   {
      return null;
   }

   public TeXObjectList expandonce(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      return null;
   }

   public TeXObjectList expandfully(TeXParser parser) throws IOException
   {
      return null;
   }

   public TeXObjectList expandfully(TeXParser parser, TeXObjectList stack)
     throws IOException
   {
      return null;
   }

   public String toString(TeXParser parser)
   {
      return toString();
   }

   public String toString()
   {
      return String.format("%c", (int)charCode);
   }

   public void process(TeXParser parser) throws IOException
   {
   }

   public void process(TeXParser parser, TeXObjectList stack) throws IOException
   {
   }

   private int charCode;
}
