/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.nio.file.*;
import java.util.*;
import java.util.zip.*;
import java.net.*;

import javax.swing.*;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import javax.help.*;

import com.dickimawbooks.texparserlib.TeXApp;
import com.dickimawbooks.texparserlib.TeXSyntaxException;
import com.dickimawbooks.texparserlib.html.L2HStringConverter;

import com.dickimawbooks.makejmlrbookgui.newbook.*;

public class MakeJmlrBookGUI extends JFrame
  implements ActionListener,MenuListener,HyperlinkListener
{
   public MakeJmlrBookGUI()
   {
      super(appName);

      initFrame();
   }

   private void initFrame()
   {
      setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);

      addWindowListener(new WindowAdapter()
      {
         public void windowClosing(WindowEvent evt)
         {
            quit();
         }
      });

      setIconImage(new ImageIcon(getClass().getResource("/icons/makejmlrbookgui-small.png")).getImage());

      boolean welcome = true;

      try
      {
         properties = JMLRBookProperties.fetchProperties(this);

         String lastVersion = properties.getVersion();

         if (appVersion.equals(lastVersion))
         {
            welcome = false;
         }
      }
      catch (IOException e)
      {
         JMLRResources.error(null, getLabelWithAlt("error.no_properties",
               "Unable to load properties")
            +"\n"+e.getMessage());
         properties = new JMLRBookProperties(this);
         e.printStackTrace();
      }

      properties.setVersion(appVersion);

      LOGWRITER = null;

      if (properties.isLogging())
      {
         File loggingFile = properties.getLoggingFile();

         try
         {
            LOGWRITER = new PrintWriter(loggingFile);
            LOGWRITER.println("Version "+appVersion);
         }
         catch (IOException e)
         {
            JMLRResources.error(null, e);
            LOGWRITER = null;
         }
      }

      toolBar = new JToolBar(properties.getToolBarOrientation());
      getContentPane().add(toolBar, properties.getToolBarPosition());

      initHelp();

      fileChooser = new JFileChooser(properties.getDefaultDirectory());
      texFileFilter = new TeXFileFilter();
      fileChooser.addChoosableFileFilter(texFileFilter);
      fileChooser.setAcceptAllFileFilterUsed(false);
      fileChooser.setFileFilter(texFileFilter);

      pdfFileChooser = new JFileChooser(properties.getDefaultDirectory());
      htmlFileChooser = new JFileChooser(properties.getDefaultDirectory());
      bibFileChooser = new JFileChooser(properties.getDefaultDirectory());
      zipFileChooser = new JFileChooser(properties.getDefaultDirectory());

      htmlFileFilter = new HtmlFileFilter();
      htmlFileChooser.addChoosableFileFilter(htmlFileFilter);
      htmlFileChooser.setFileFilter(htmlFileFilter);

      bibFileFilter = new BibFileFilter();
      bibFileChooser.addChoosableFileFilter(bibFileFilter);
      bibFileChooser.setFileFilter(bibFileFilter);

      zipFileFilter = new ZipFileFilter();
      zipFileChooser.addChoosableFileFilter(zipFileFilter);
      zipFileChooser.setFileFilter(zipFileFilter);

      imageFileFilter = new ImageFileFilter();
      bookContentsFilter = new BookContentsFilter();

      dirChooser = new JFileChooser(properties.getDefaultDirectory());
      dirChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

      appSelector = new JMLRAppSelector(this);

      JMenuBar mBar = new JMenuBar();
      setJMenuBar(mBar);

      JMenu fileM = createMenu("file");

      mBar.add(fileM);

      newItem = createMenuButtonItem("file", "new",
         KeyStroke.getKeyStroke(KeyEvent.VK_N, InputEvent.CTRL_MASK),
         "general/New24", fileM);

      newBookDialog = new NewBookDialog(this);

      recentFilesListener = new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            String action = evt.getActionCommand();

            if (action == null) return;

            try
            {
               int idx = Integer.parseInt(action);

               File file = new File(properties.getRecentFileName(idx));

               fileChooser.setCurrentDirectory(file.getParentFile());

               load(file);
            }
            catch (IOException e)
            {
               error(e);
            }
            catch (NumberFormatException e)
            {
            }
         }
      };

      recentM = createMenu("file", "recent");
      recentM.addMenuListener(this);
      fileM.add(recentM);

      clearRecentItem = createMenuItem("file.recent", "clearrecent");

      openItem = createMenuButtonItem("file", "open",
         KeyStroke.getKeyStroke(KeyEvent.VK_O, InputEvent.CTRL_MASK),
         "general/Open24", fileM);

      exportItem = createMenuButtonItem("file", "export",
         KeyStroke.getKeyStroke(KeyEvent.VK_E, InputEvent.CTRL_MASK),
         "general/Export24", fileM);

      fileM.addSeparator();

      clearItem = createMenuItem("file", "clear",
         KeyStroke.getKeyStroke(KeyEvent.VK_W, InputEvent.CTRL_MASK));
      fileM.add(clearItem);

      JMenuItem quitItem = createMenuItem("file", "quit",
         KeyStroke.getKeyStroke(KeyEvent.VK_Q, InputEvent.CTRL_MASK));

      fileM.add(quitItem);

      toolBar.addSeparator();
      buildM = createMenu("build");
      mBar.add(buildM);

      buildMainItem = createMenuButtonItem("build", "buildbook",
         KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK),
         "book.png", buildM);

      buildPrintItem = createMenuButtonItem("build", "buildprint",
         KeyStroke.getKeyStroke(KeyEvent.VK_P, InputEvent.CTRL_MASK),
         "book-print.png", buildM);

      buildPapersM = createMenu("build", "papers");
      buildPapersM.addMenuListener(this);

      buildM.add(buildPapersM);

      createBibItem = createMenuItem("build", "createbib");
      buildM.add(createBibItem);

      createZipItem = createMenuItem("build", "createzip");
      buildM.add(createZipItem);

      toolBar.addSeparator();
      toolsM = createMenu("tools");

      mBar.add(toolsM);

      refreshItem = createMenuButtonItem("tools", "refresh",
         KeyStroke.getKeyStroke(KeyEvent.VK_R, InputEvent.CTRL_MASK),
         "general/Refresh24", toolsM);
      refreshItem.setEnabled(false);

      removeAuxItem = createMenuButtonItem("tools", "removeaux",
         KeyStroke.getKeyStroke(KeyEvent.VK_X, InputEvent.CTRL_MASK),
        "general/Delete24", toolsM);

      toolsM.add(createMenuItem("tools", "fetch_jmlr_cls",
        KeyStroke.getKeyStroke(KeyEvent.VK_J, InputEvent.CTRL_MASK)));

      if (debugMode)
      {
         pdfxItem = createMenuItem("tools", "pdfx", 
            KeyStroke.getKeyStroke(KeyEvent.VK_T, InputEvent.CTRL_MASK));
         toolsM.add(pdfxItem);
      }

      charCheckItem = createMenuItem("tools", "charcheck");
      toolsM.add(charCheckItem);

      charCheckSelector = new CharCheckSelector(this);

      toolBar.addSeparator();

      abortItem = createMenuButtonItem("tools", "abort",
         KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0),
         "general/Stop24", toolsM);

      abortItem.setEnabled(false);

      toolBar.addSeparator();

      JMenu settingsM = createMenu("settings");

      mBar.add(settingsM);

      createMenuButtonItem("settings", "editsettings",
         KeyStroke.getKeyStroke(KeyEvent.VK_S, InputEvent.CTRL_MASK),
         "general/Preferences24", settingsM);

      propertiesDialog = new PropertiesDialog(this);

      JMenu helpM = createMenu("help");
      mBar.add(helpM);

      helpM.add(createMenuItem("help", "about"));

      JMenuItem item = new JMenuItem(getLabel("help",
        "license"));
      item.setMnemonic(getMnemonic("help", "license"));

      enableHelpOnButton(item, "licence");

      helpM.add(item);

      createMenuButtonItem("help", "manual",
         KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0),
         "general/Help24", csh, helpM);

      tabbedPane = new JTabbedPane();

      mainPanel = new JMLRBookPanel(this);
      tabbedPane.add(mainPanel);

      fileViewer = new FileViewer(this);
      tabbedPane.add(getLabel("fileview.title"), new JScrollPane(fileViewer));

      messageArea = new MessageLabel(getLabel("message.init"));

      JPanel panel = new JPanel(new BorderLayout());

      panel.add(tabbedPane, "Center");
      panel.add(messageArea, "South");

      getContentPane().add(panel);

      currentProcessListeners = new Vector<ProcessListener>();
      currentWorkerThreads = new Vector<Thread>();
      kpsewhichResults = new HashMap<String,String>();

      clear();

      Toolkit tk = Toolkit.getDefaultToolkit();
      Dimension d = tk.getScreenSize();
      setSize(d.width*3/4, d.height*3/4);

      setLocationRelativeTo(null);

      l2hStringConverter = new L2HStringConverter(new BookTeXApp(this))
      {
         protected void addPredefined()
         {
            super.addPredefined();
            putControlSequence(new L2HName());
            putControlSequence(new L2HEmail());
            putControlSequence(new L2HAnd());
            putControlSequence(new L2HAnd("And"));
            putControlSequence(new L2HAddr());
            putControlSequence(new L2HInputEncoding());
         }
      };

      informationDialog = new InformationDialog(this);

      if (welcome)
      {
         SwingUtilities.invokeLater(new Runnable()
         {
            public void run()
            {
               welcome();
            }
         });
      }
   }

   private void welcome()
   {
      (new WelcomeDialog(this)).setVisible(true);
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("quit"))
      {
         quit();
      }
      else if (action.equals("new"))
      {
         newBookDialog.display();
      }
      else if (action.equals("editsettings"))
      {
         propertiesDialog.display();
      }
      else if (action.equals("clear"))
      {
         clear();
      }
      else if (action.equals("open"))
      {
         if (fileChooser.showOpenDialog(this)
           == JFileChooser.APPROVE_OPTION)
         {
            try
            {
               load(fileChooser.getSelectedFile());
            }
            catch (IOException e)
            {
               error(e);
            }
         }
      }
      else if (action.equals("export"))
      {
         if (htmlFileChooser.showOpenDialog(this)
           == JFileChooser.APPROVE_OPTION)
         {
            try
            {
               File file = htmlFileChooser.getSelectedFile();

               if (!file.getName().contains(".")
                && htmlFileChooser.getFileFilter() instanceof HtmlFileFilter)
               {
                  file = new File(file.getParent(), file.getName()+".html");
               }

               mainPanel.export(file);
            }
            catch (IOException e)
            {
               error(e);
            }
         }
      }
      else if (action.equals("buildbook"))
      {
         Thread thread = new ProcessThread(this, getBookRoot(), false, false, true);
         startWorkerThread(thread);
      }
      else if (action.equals("buildprint"))
      {
         Thread thread = new ProcessThread(this, getBookRoot(), false, true, true);
         startWorkerThread(thread);
      }
      else if (action.equals("createbib"))
      {
         try
         {
            createBib();
         }
         catch (IOException e)
         {
            error(e);
         }
      }
      else if (action.equals("createzip"))
      {
         try
         {
            createZip();
         }
         catch (IOException e)
         {
            error(e);
         }
      }
      else if (action.equals("removeaux"))
      {
         kpsewhichResults.clear();
         Thread thread = new ClearAuxThread(this, getBook());

         thread.start();
         thread = null;
      }
      else if (action.equals("clearrecent"))
      {
         properties.clearRecentList();
      }
      else if (action.equals("refresh"))
      {
         Thread thread = new Thread()
         {
            public void run()
            {
               disableFunctions();
               setWaitCursor();
               refreshBook();
               setDefaultCursor();
               resetFunctions();
            }
         };
         thread.start();
         thread = null;
      }
      else if (action.equals("fetch_jmlr_cls"))
      {
         Thread thread = new FetchClsThread(this);

         startWorkerThread(thread);
      }
      else if (action.equals("pdfx"))
      {
         if (pdfFileChooser.showOpenDialog(this) == JFileChooser.APPROVE_OPTION)
         {
            Thread thread = new PdfXThread(this,
               pdfFileChooser.getSelectedFile());
            startWorkerThread(thread);
         }
      }
      else if (action.equals("charcheck"))
      {
         try
         {
            BookRoot selection = charCheckSelector.requestFiles(bookRoot);

            if (selection != null)
            {
               Thread thread = new CharCheckThread(this, selection);

               startWorkerThread(thread);
            }
         }
         catch (IOException e)
         {
            error(e);
         }
      }
      else if (action.equals("abort"))
      {
         abort();
      }
      else if (action.equals("about"))
      {
         String[] str;

         String translator = dictionary.getProperty("about.translator_info");

         if (translator == null || translator.equals(""))
         {
            str = new String[]
            {
               appName,
               getLabelWithValue("about.version", appVersion),
               getLabelWithValues("about.copyright", "Nicola L. C. Talbot",
                   appDate),
               "http://www.dickimaw-books.com/"
            };
         }
         else
         {
            str = new String[]
            {
               appName,
               getLabelWithValue("about.version", appVersion),
               getLabelWithValues("about.copyright", "Nicola L. C. Talbot",
                  appDate),
               "http://www.dickimaw-books.com/",
               translator
            };
         }

         JOptionPane.showMessageDialog(this, str,
            getLabelWithValue("about.title", appName), JOptionPane.PLAIN_MESSAGE);
      }
   }

   public void hyperlinkUpdate(HyperlinkEvent evt)
   {
      HyperlinkEvent.EventType type = evt.getEventType();

      if (type == HyperlinkEvent.EventType.ACTIVATED)
      {
         URL url = evt.getURL();

         String description = evt.getDescription();

         if (description != null)
         {
            if (description.startsWith("#"))
            {
               mainPanel.scrollToReference(description.substring(1));
               return;
            }
            else if (description.startsWith("info:"))
            {
               String info = description.substring(description.indexOf(":")+1);

               int idx = info.indexOf(":");

               if (idx == -1)
               {
                  informationDialog.display(info);
               }
               else
               {
                  String label = info.substring(0, idx);
                  String rest = info.substring(idx+1);

                  informationDialog.display(label, rest.split(":"));
               }

               return;
            }
         }

         if (url != null)
         {
            viewFile(url);
         }

         MakeJmlrBookGUI.debug("Don't know what to do with hyperlink update for event "+evt);
      }
   }

   public void menuSelected(MenuEvent evt)
   {
      Object source = evt.getSource();

      if (source == recentM)
      {
         properties.setRecentFiles(recentM, recentFilesListener);

         recentM.addSeparator();
         recentM.add(clearRecentItem);
      }
      else if (source == buildPapersM)
      {
         if (bookRoot != null)
         {
            Book book = bookRoot.getBook();

            if (book != null)
            {
               book.buildPapersMenu(buildPapersM);
            }
         }
      }
   }

   public void menuDeselected(MenuEvent evt)
   {
   }

   public void menuCanceled(MenuEvent evt)
   {
   }

   public void quit()
   {
      try
      {
         properties.setToolBarPosition((String)
            ((BorderLayout)getContentPane().getLayout()).getConstraints(toolBar));
         properties.setToolBarOrientation(toolBar.getOrientation());
         properties.save();
      }
      catch (Exception e)
      {
         error(getLabel("error.propsave_failed")+"\n"+e.getMessage());
         e.printStackTrace();
      }

      if (LOGWRITER != null)
      {
         LOGWRITER.close();
      }

      System.exit(0);
   }


   public void clear()
   {
      bookRoot = null;
      fileViewer.updateBook();
      setTitle(appName);

      openItem.setEnabled(true);
      exportItem.setEnabled(false);
      newItem.setEnabled(true);
      recentM.setEnabled(true);
      mainPanel.clear();
      disableTools();

      kpsewhichResults.clear();
   }

   public void createZip() throws IOException
   {
      Book book = getBook();

      if (book == null)
      {
         throw new IOException("no book found!");
      }

      File file = zipFileChooser.getSelectedFile();

      if (file == null)
      {
         String title = book.getHtmlShortTitle().replaceAll("[^\\w]+", "");
         zipFileChooser.setCurrentDirectory(book.getLocation());
         zipFileChooser.setSelectedFile(new File(title+".zip"));
      }

      if (zipFileChooser.showSaveDialog(this)
        != JFileChooser.APPROVE_OPTION)
      {
         return;
      }

      file = zipFileChooser.getSelectedFile();

      if (file.exists())
      {
         if (JOptionPane.showConfirmDialog(this, 
               getLabelWithValue("message.confirm_overwrite", 
                                 file.getAbsolutePath()),
               getLabel("message.confirm"),
               JOptionPane.YES_NO_OPTION,
               JOptionPane.QUESTION_MESSAGE)
           != JOptionPane.YES_OPTION)
         {
            return;
         }
      }

      book.compress(file);
      setInfo(getLabel("message.done"));
   }

   public BibFileFilter getBibFileFilter()
   {
      return bibFileFilter;
   }

   public JFileChooser getBibFileChooser()
   {
      return bibFileChooser;
   }

   public void createBib()
    throws IOException
   {
      Book book = getBook();

      if (book == null)
      {
         throw new IOException("no book found!");
      }

      File file = bibFileChooser.getSelectedFile();

      if (file == null)
      {
         String title = book.getHtmlShortTitle().replaceAll("[^\\w]+", "");
         bibFileChooser.setCurrentDirectory(book.getLocation());
         bibFileChooser.setSelectedFile(new File(title+".bib"));
      }

      if (bibFileChooser.showSaveDialog(this)
        != JFileChooser.APPROVE_OPTION)
      {
         return;
      }

      file = bibFileChooser.getSelectedFile();

      if (file.exists())
      {
         if (JOptionPane.showConfirmDialog(this, 
               getLabelWithValue("message.confirm_overwrite", 
                                 file.getAbsolutePath()),
               getLabel("message.confirm"),
               JOptionPane.YES_NO_OPTION,
               JOptionPane.QUESTION_MESSAGE)
           != JOptionPane.YES_OPTION)
         {
            return;
         }
      }

      book.createBib(file);
      setInfo(getLabel("message.done"));
   }

   public void load(File file)
    throws IOException
   {
      // check file exists

      if (!file.exists())
      {
         error(getLabelWithValue("error.file_not_found", file.getAbsolutePath()));

         return;
      }

      properties.addRecentFile(file.getAbsolutePath());

      debug("Loading "+file);

      Book b = new Book(this, file);

      setBookRoot(b.getRoot(), b.isRunRequired());
   }

   public boolean refreshBook()
   {
      setInfo(getLabel("message.updating"));

      BookRoot root = getBookRoot();

      if (root == null)
      {
         return false;
      }

      getBook().populate();

      fileViewer.updateBook();
      updateMainPanel();
      setInfo(getLabel("message.done"));

      return true;
   }


   public void setWaitCursor()
   {
      Thread thread = new Thread()
      {
         public void run()
         {
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
         }
      };
      thread.start();
      thread = null;
   }

   public void setDefaultCursor()
   {
      Thread thread = new Thread()
      {
         public void run()
         {
            setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
         }
      };
      thread.start();
      thread = null;
   }

   public void disableTools()
   {
      debug("disabling tools");
      setInfo("disabling tools");
      buildM.setEnabled(false);
      buildMainItem.setEnabled(false);
      buildPrintItem.setEnabled(false);
      createBibItem.setEnabled(false);
      createZipItem.setEnabled(false);
      removeAuxItem.setEnabled(false);
      refreshItem.setEnabled(false);
      clearItem.setEnabled(false);
      charCheckItem.setEnabled(false);
      exportItem.setEnabled(false);
      setInfo(getLabel("message.done"));
   }

   public void enableTools()
   {
      buildM.setEnabled(true);
      buildMainItem.setEnabled(true);
      buildPrintItem.setEnabled(true);
      createBibItem.setEnabled(true);
      createZipItem.setEnabled(true);
      refreshItem.setEnabled(true);
      removeAuxItem.setEnabled(true);
      clearItem.setEnabled(true);
      abortItem.setEnabled(false);
      charCheckItem.setEnabled(true);
      exportItem.setEnabled(true);
   }

   public void disableFunctions()
   {
      openItem.setEnabled(false);
      exportItem.setEnabled(false);
      newItem.setEnabled(false);
      recentM.setEnabled(false);
      disableTools();
   }

   public void resetFunctions()
   {
      if (bookRoot == null)
      {
         openItem.setEnabled(true);
         exportItem.setEnabled(false);
         newItem.setEnabled(true);
         recentM.setEnabled(true);

         disableTools();
      }
      else
      {
         openItem.setEnabled(false);
         exportItem.setEnabled(true);
         newItem.setEnabled(false);
         recentM.setEnabled(false);

         enableTools();
      }
   }

   public Book getBook()
   {
      if (bookRoot == null)
      {
         return null;
      }

      return bookRoot.getBook();
   }

   public BookRoot getBookRoot()
   {
      return bookRoot;
   }

   public void setBookRoot(BookRoot root)
   {
      setBookRoot(root, false);
   }

   public void setBookRoot(BookRoot root, boolean dorun)
   {
      bookRoot = root;
      setTitle(appName+" - "+bookRoot.getName());

      openItem.setEnabled(false);
      exportItem.setEnabled(true);
      newItem.setEnabled(false);
      recentM.setEnabled(false);
      enableTools();

      buildMainItem.setText(getLabel("build.buildbook"));

      buildPapersM.setEnabled(true);
      buildPrintItem.setEnabled(true);

      refreshItem.setEnabled(true);

      debug("Updating fileviewer");
      fileViewer.updateBook();

      try
      {
         debug("Checking for latex app");
         getLaTeXApp();
      }
      catch (FileNotFoundException e)
      {
         warning(e.getMessage());
      }

      try
      {
         debug("Checking for bibtex app");
         getBibTeXApp();
      }
      catch (FileNotFoundException e)
      {
         warning(e.getMessage());
      }

      debug("Spawning process");
      Thread thread = new ProcessThread(this, bookRoot, false, false, dorun);
      debug("Starting worker thread");
      startWorkerThread(thread);
   }

   public ImageFileFilter getImageFileFilter()
   {
      return imageFileFilter;
   }

   public BookContentsFilter getBookContentsFilter()
   {
      return bookContentsFilter;
   }

   public JFileChooser getFileChooser()
   {
      return fileChooser;
   }

   public JFileChooser getDirChooser()
   {
      return dirChooser;
   }

   public String getCurrentDirectory()
   {
      return fileChooser.getCurrentDirectory().getAbsolutePath();
   }

   public JMLRBookProperties getProperties()
   {
      return properties;
   }

   public boolean isStrict()
   {
      return properties.isStrict();
   }

   public boolean showDetailsIcon()
   {
      return properties.showDetailsIcon();
   }

   public boolean showDetailsText()
   {
      return properties.showDetailsText();
   }

   public Book requestMainFile(BookRoot root)
     throws IOException
   {
      JTree tree = new JTree(root);
      tree.getSelectionModel()
         .setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

      JPanel panel = new JPanel(new BorderLayout());

      panel.add(new JLabel(getLabel("error.cant_find_main")), "North");
      panel.add(new JScrollPane(tree), "Center");

      if (JOptionPane.showConfirmDialog(this, panel,
           getLabel("file.select"),
           JOptionPane.OK_CANCEL_OPTION)
          != JOptionPane.OK_OPTION)
      {
         return null;
      }

      TreePath path = tree.getSelectionPath();

      if (path == null)
      {
         return null;
      }

      FileElement element = (FileElement)path.getLastPathComponent();

      if (element.isDirectory())
      {
         error(getLabelWithValue("error.dir.is", element.getName()));

         return requestMainFile(root);
      }

      return root.setMainFile((BookFile)element);
   }

   public void bookLoaded()
   {
      mainPanel.bookLoaded();
   }

   public void updateMainPanel()
   {
      mainPanel.updateInfo();
   }

   public void updateFileView()
   {
      fileViewer.updateBook();
   }

   public void refreshFileView()
   {
      fileViewer.refresh();
   }

   public L2HStringConverter getL2HStringConverter()
   {
      return l2hStringConverter;
   }

   public String toHtml(String texSource, boolean atIsLetter)
   {
      try
      {
         return l2hStringConverter.convert(texSource, atIsLetter);
      }
      catch (IOException e)
      {
      }

      return texSource;
   }

   public void message(String msg)
   {
      setInfoAndDebug(msg);
   }

   public String getMessage(String label)
   {
      return getLabel(label);
   }

   public String getMessage(String label, String param)
   {
      return getLabelWithValue(label, param);
   }

   public String getMessage(String label, String[] params)
   {
      return getLabelWithValues(label, params);
   }

   public void substituting(int lineNum, String original, String replacement)
   {
      if (replacement.isEmpty())
      {
         setInfoAndDebug(getLabelWithValues("error.line", ""+lineNum,
           getLabelWithValue("warning.removing", tag(original))));
      }
      else
      {
         setInfoAndDebug(getLabelWithValues("error.line", ""+lineNum,
           getLabelWithValues("warning.substituting",
              tag(original), tag(replacement))));
      }
   }

   public String tag(String string)
   {
      return string;
   }

   public void setInfo(String message)
   {
      if (messageArea != null)
      {
         messageArea.update(message);
      }
   }

   public void setInfoAndDebug(String message)
   {
      debug(message);
      setInfo(message);
   }

   public void error(String message)
   {
      JMLRResources.error(this, message);
   }

   public String getErrorMessage(TeXSyntaxException excp)
   {
      String[] params = excp.getParams();
      File file = excp.getFile();
      int lineNum = excp.getLineNumber();
      String errorTag = excp.getErrorTag();

      String msg =
        params == null ? getMessage(errorTag) : getMessage(errorTag, params);

      if (file == null && lineNum == -1)
      {
         return msg;
      }
      else if (lineNum == -1)
      {
         return String.format("%s: %s", file.getName(), msg);
      }
      else if (file == null)
      {
         return String.format("l.%d: %s", lineNum, msg);
      }

      return String.format("%s:%d: %s", file.getName(), lineNum, msg);
   }

   public void error(Exception excp)
   {
      if (excp instanceof TeXSyntaxException)
      {
          JMLRResources.error(this, 
            getErrorMessage((TeXSyntaxException)excp), excp);
      }
      else
      {
         JMLRResources.error(this, excp);
      }
   }

   public void warning(String message)
   {
      JMLRResources.warning(this, message);
   }

   public void copyFile(File src, File dest)
   throws IOException
   {
      verboseCopyFile(src, dest);
   }

   public static void copyFile(File src, String destDir, String destName)
   throws IOException
   {
      copyFile(src.getParent(), src.getName(), destDir, destName);
   }

   public static void copyFile(String srcDir, String srcName, 
      String destDir, String destName)
   throws IOException
   {
      Path source = FileSystems.getDefault().getPath(
         srcDir, srcName);
      Path target = FileSystems.getDefault().getPath(
         destDir, destName);

      File destDirFile = new File(destDir);

      if (!destDirFile.exists())
      {
          Files.createDirectories(destDirFile.toPath());
      }

      Files.copy(source, target, StandardCopyOption.REPLACE_EXISTING);
   }

   public void verboseCopyFile(File src, File dest)
   throws IOException
   {
      verboseCopyFile(src.getParent(), src.getName(),
               dest.getParent(), dest.getName());
   }

   public void verboseCopyFile(File src, String destDir, String destName)
   throws IOException
   {
      verboseCopyFile(src.getParent(), src.getName(), destDir, destName);
   }

   public void verboseCopyFile(String srcDir, String srcName, 
      String destDir, String destName)
   throws IOException
   {
      Path source = FileSystems.getDefault().getPath(
         srcDir, srcName);
      Path target = FileSystems.getDefault().getPath(
         destDir, destName);

      File destDirFile = new File(destDir);

      if (!destDirFile.exists())
      {
          Files.createDirectories(destDirFile.toPath());
      }

      String mess = getLabelWithValues("message.copying",
         source.toString(), target.toString());
      debug(mess);
      setInfo(mess);

      Files.copy(source, target, StandardCopyOption.REPLACE_EXISTING);
   }

   public void verboseMoveFile(File srcFile, File destFile)
     throws IOException
   {
      Path srcPath = srcFile.toPath();
      Path destPath = destFile.toPath();

      String mess = getLabelWithValues("message.moving",
         srcPath.toString(), destPath.toString());
      debug(mess);
      setInfo(mess);

      Files.move(srcPath, destPath);
   }

   public static boolean ctanPkgExists(String name)
    throws IOException
   {
      URL url = new URL("http://ctan.org/pkg/"+name);
      debug("Checking the existence of "+url);
      HttpURLConnection con = (HttpURLConnection)url.openConnection();
      con.setRequestMethod("HEAD");
      con.connect();
      int code = con.getResponseCode();

      debug("Return code: "+code);

      return code == HttpURLConnection.HTTP_OK;
   }

   public void beepIfEnabled()
   {
      if (properties.isBeepEnabled())
      {
         java.awt.Toolkit.getDefaultToolkit().beep();
      }
   }

   public void viewFile(File file)
      throws MalformedURLException
   {
      viewFile(file.toURI().toURL());
   }

   public void viewFile(URL url)
   {
      String file = url.getFile();
      String fileLocation = url.toString();

      String viewer = null;

      if (file.endsWith(".pdf"))
      {
         viewer = properties.getPdfViewer();

         if (viewer == null || viewer.equals(""))
         {
            viewer = appSelector.fetchApplicationPath(
                getLabelWithValue("properties.query.location.viewer", "PDF"))
                   .getAbsolutePath();

            if (viewer == null || viewer.equals(""))
            {
               error(getLabelWithValue("error.load_type_failed", "PDF"));
               return;
            }

            properties.setPdfViewer(viewer);
         }
      }
      else if (url.getProtocol().startsWith("http") ||
               url.getProtocol().startsWith("mailto") ||
          file.endsWith(".html") || file.endsWith(".htm"))
      {
         viewer = properties.getHtmlViewer();

         if (viewer == null || viewer.equals(""))
         {
            viewer = appSelector.fetchApplicationPath(
                getLabelWithValue("properties.query.location.viewer", "HTML"))
                   .getAbsolutePath();

            if (viewer == null || viewer.equals(""))
            {
               error(getLabelWithValue("error.load_type_failed", "HTML"));
               return;
            }

            properties.setHtmlViewer(viewer);
         }
      }
      else if (imageFileFilter.accept(new File(file)))
      {
         viewer = properties.getImageViewer();

         if (viewer == null || viewer.equals(""))
         {
            viewer = appSelector.fetchApplicationPath(
               getLabel("properties.query.location.imageviewer")).getAbsolutePath();

            if (viewer == null || viewer.equals(""))
            {
               error(getLabel("error.load_failed"));
               return;
            }

            properties.setImageViewer(viewer);
         }

         fileLocation = file;
      }
      else
      {
         viewer = properties.getTextViewer();

         if (viewer == null || viewer.equals(""))
         {
            File app = appSelector.findApp("texworks");

            if (app == null)
            {
               viewer = appSelector.fetchApplicationPath(
                  getLabel("properties.query.location.texteditor"))
                    .getAbsolutePath();

               if (viewer == null || viewer.equals(""))
               {
                  error(getLabel("error.load_failed"));
                  return;
               }
            }
            else
            {
               viewer = app.getAbsolutePath();
            }

            properties.setTextViewer(viewer);
         }

         fileLocation = file;
      }


      String[] cmd = new String[]{viewer, fileLocation};

      try
      {
         Process p = execCommand(cmd, false);
      }
      catch (Exception e)
      {
         error(getLabelWithValue("error.exec_failed", cmd[0]+" "+cmd[1])
            +"\n"+e.getMessage());
      }
   }

   public File getICCFile()
     throws IOException,InterruptedException
   {
      String iccPath = properties.getICCFile();
      File file = null;

      if (iccPath ==  null || iccPath.length() == 0)
      {
         file = null;
      }
      else
      {
         file = new File(iccPath);
      }

      if (file != null)
      {
         if (file.exists())
         {
            return file;
         }
         else
         {
            file = null;
         }
      }

      FetchICCProcessListener listener = new FetchICCProcessListener(this);

      int exitCode = execCommandAndWaitFor(new String[]{"kpsewhich", "FOGRA39L.icc"}, listener);

      if (exitCode != 0)
      {
         error(getLabel("error.checking_stderr")+"\n"+listener.getErrors());
      }

      file = listener.getFile();

      if (file == null)
      {
         if (JOptionPane.showConfirmDialog(this, getLabel("error.missing_icc"),
            getLabel("error.title"), JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE)
            == JOptionPane.YES_OPTION)
         {
            URL url = new URL("http://www.colormanagement.org/download_files/ISOcoated_v2_basICC.zip");

            File targetDir = getTeXMF();

            // Confirm destination directory

            JFileChooser dirChooser = new JFileChooser(
               targetDir == null ? null : targetDir.getParentFile());
            dirChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

            if (targetDir != null)
            {
               dirChooser.setSelectedFile(targetDir);
            }

            if (dirChooser.showDialog(this, getLabel("button.select"))
                != JFileChooser.APPROVE_OPTION)
            {
               return null;
            }

            targetDir = dirChooser.getSelectedFile();

            unpackZipEntry(url, targetDir, ".icc", "FOGRA39L.icc");

            String[] cmd = 
               new String[]
               {
                  properties.getTeXHash(),
                  targetDir.getAbsolutePath()
               };

            try
            {
               exitCode = execCommandAndWaitFor(cmd);

               if (exitCode != 0)
               {
                  throw new ProcessFailedException(cmd[0]+" \""+cmd[1]+"\"");
               }
            }
            catch (Exception e)
            {
               error(e);
            }
         }
      }

      if (file != null)
      {
         properties.setICCFile(file.getAbsolutePath());
      }

      return file;
   }

   public int getPdfPageCount(File file)
     throws IOException,InterruptedException
   {
      String gsApp = getGSApp();
      int pageCount = -1;

      BufferedReader in = null;

      try
      {
         String[] cmd = new String[]
         {
            gsApp,
            "-q",
            "-dNODISPLAY",
            "-c",
            "("+file.getAbsolutePath()+") (r) file runpdfbegin pdfpagecount = quit"
         };

         String info = gsApp;

         for (int i = 1; i < cmd.length; i++)
         {
            info += " "+cmd[i];
         }

         setInfo(getLabelWithValue("message.running", info));

         PageCountListener listener = new PageCountListener(this);

         int exitCode = execCommandAndWaitFor(cmd, listener);

         pageCount = listener.getPageCount();
      }
      finally
      {
         if (in != null)
         {
            in.close();
            in = null;
         }
      }

      return pageCount;
   }

   public void toPdfX(File pdfFile)
      throws IOException,InterruptedException
   {
      toPdfX(null, null, null, getICCFile(), pdfFile);
   }

   public void toPdfX(String author, String title, String info, File iccFile, File pdfFile)
      throws IOException,InterruptedException
   {
      String gsApp = getGSApp();

      // create def file

      File dir = pdfFile.getParentFile();

      File defFile = new File(dir, "PDFX_def.ps");

      PrintWriter out = null;

      setInfo(getLabelWithValue("message.writing", defFile.getAbsolutePath()));

      try
      {
         out = new PrintWriter(new FileWriter(defFile));

         out.println("systemdict /ProcessColorModel known {");
         out.println("systemdict /ProcessColorModel get dup /DeviceGray ne exch /DeviceCMYK ne and } {");
         out.println("  true");
         out.println("} ifelse");
         out.println("{ (ERROR: ProcessColorModel must be /DeviceGray or DeviceCMYK.)=");
         out.println("  /ProcessColorModel cvx /rangecheck signalerror");
         out.println("} if");

         out.println();
         out.println("% Define entries to the document Info dictionary :");
         out.println();

         out.println("/ICCProfile ("+iccFile.getAbsolutePath()+") def");
         out.println();

         out.println("[ /GTS_PDFXVersion (PDF/X-3:2002)");

         if (title != null)
         {
            out.println("  /Title ("+title.replaceAll("\\(", "\\\\(").replaceAll("\\)", "\\\\)")+")");
         }

         if (author != null)
         {
            out.println("  /Author ("+author.replaceAll("\\(", "\\\\(").replaceAll("\\)", "\\\\)")+")");
         }

         out.println("  /Trapped /False");
         out.println("  /DOCINFO pdfmark");

         out.println();
         out.println("% Define an ICC profile :");
         out.println();

         out.println("currentdict /ICCProfile known {");
         out.println("  [/_objdef {icc_PDFX} /type /stream /OBJ pdfmark");
         out.println("  [{icc_PDFX} <</N systemdict /ProcessColorModel get /DeviceGray eq {1} {4} ifelse >> /PUT pdfmark");
         out.println("  [{icc_PDFX} ICCProfile (r) file /PUT pdfmark");
         out.println("} if");

         out.println();
         out.println("% Define the output intent dictionary :");
         out.println();

         out.println("[/_objdef {OutputIntent_PDFX} /type /dict /OBJ pdfmark");
         out.println("[{OutputIntent_PDFX} <<");
         out.println("  /Type /OutputIntent");
         out.println("  /S /GTS_PDFX");
         out.println("  /OutputCondition (Commercial and speciality printing)");

         if (info != null)
         {
            out.println("  /Info ("+info.replaceAll("\\(", "\\\\(").replaceAll("\\)", "\\\\)")+")");
         }

         out.println("  /OutputConditionIdentifier (FOGRA39)");
         out.println("  /RegistryName (http://www.color.org)");
         out.println("  currentdict /ICCProfile known {");
         out.println("    /DestOutputProfile {icc_PDFX}");
         out.println("  } if");
         out.println(">> /PUT pdfmark");
         out.println("[{Catalog} <</OutputIntents [ {OutputIntent_PDFX} ]>> /PUT pdfmark");

         File tmpFile = File.createTempFile("ciml", ".pdf");
         tmpFile.deleteOnExit();

         String[] cmd = 
            new String[]
            {
               gsApp,
               "-dPDFX",
               "-dBATCH",
               "-dNOPAUSE",
               "-dNOOUTERSAVE",
               "-sDEVICE=pdfwrite",
               "-sColorConversionStrategy=Gray",
               "-dProcessColorModel=/DeviceGray",
               "-dPDFSETTINGS=/prepress",
               "-sOutputFile="+tmpFile.getAbsolutePath()+"",
               defFile.getAbsolutePath(),
               pdfFile.getAbsolutePath()
            };

         PdfxProcessListener listener = new PdfxProcessListener(this);

         int exitCode = execCommandAndWaitFor(cmd, listener);

         if (exitCode != 0)
         {
            String mess = getLabelWithValue("error.exec_failed", info+listener.getErrors());
            error(mess);
            System.err.println(mess);

            return;
         }

         String mess = listener.getMessage();

         if (mess != null)
         {
            warning(mess);
         }

         verboseCopyFile(tmpFile, pdfFile);
      }
      finally
      {
         if (out != null)
         {
            out.close();

            out = null;
         }
      } 

      beepIfEnabled();
   }

   public String getLaTeXApp()
      throws FileNotFoundException
   {
      String app = properties.getLaTeXApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = null;

      try
      {
         path = getApp("pdflatex");
      }
      catch (FileNotFoundException e)
      {
         disableTools();
         throw(e);
      }

      if (path != null)
      {
         properties.setLaTeXApp(path.getAbsolutePath());
      }

      return path.getAbsolutePath();
   }

   public String getBibTeXApp()
      throws FileNotFoundException
   {
      String app = properties.getBibTeXApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = getApp("bibtex");

      if (path != null)
      {
         properties.setBibTeXApp(path.getAbsolutePath());
      }

      return path.getAbsolutePath();
   }

   public File getApp(String appName)
      throws FileNotFoundException
   {
      return getApp(appName, null);
   }

   public File getApp(String appName, String altAppName)
      throws FileNotFoundException
   {
      return getApp(appName, altAppName, null);
   }

   public File getApp(String appName, String altAppName, String altAppName2)
      throws FileNotFoundException
   {
      File path = appSelector.fetchApplicationPath(
         appName, altAppName, altAppName2,
         getLabelWithValue("properties.query.location.app", appName));

      if (path == null || !path.exists())
      {
         String mess = getLabelWithValue("warning.no_build.missing.app", appName);

         throw new FileNotFoundException(mess);
      }

      return path;
   }


   public String getGSApp()
    throws IOException
   {
      String app = properties.getGSApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = null;

      try
      {
         path = getApp("gs", "gswin32c", "gsos2");
      }
      catch (FileNotFoundException e)
      {
         disableTools();
         throw(e);
      }

      String absPath = path.getAbsolutePath();

      if (path != null)
      {
         properties.setGSApp(absPath);
      }

      return absPath;
   }

   public String getKpsewhichApp()
     throws IOException
   {
      return getKpsewhichApp(true);
   }

   public String getKpsewhichApp(boolean promptIfNotFound)
     throws IOException
   {
      String app = properties.getKpsewhichApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = null;

      if (promptIfNotFound)
      {
         path = getApp("kpsewhich");
      }
      else
      {
         path = appSelector.findApp("kpsewhich");
      }

      if (path == null)
      {
         throw new FileNotFoundException(
            getLabelWithValue("error.file_cant_find", "kpsewhich"));
      }

      String absPath = path.getAbsolutePath();

      if (path != null)
      {
         properties.setKpsewhichApp(absPath);
      }

      return absPath;
   }

   public String getEpstopdfApp()
     throws IOException
   {
      String app = properties.getEpstopdfApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = getApp("epstopdf");

      String absPath = path.getAbsolutePath();

      if (path != null)
      {
         properties.setEpstopdfApp(absPath);
      }

      return absPath;
   }

   public String getWmftoepsApp()
     throws IOException
   {
      String app = properties.getWmftoepsApp();

      if (app != null && app.length() > 0)
      {
         return app;
      }

      File path = getApp("wmftoeps");

      String absPath = path.getAbsolutePath();

      if (path != null)
      {
         properties.setWmftoepsApp(absPath);
      }

      return absPath;
   }

   public void epstopdf(File file)
     throws IOException,InterruptedException
   {
      String name = file.getName();

      int idx = name.lastIndexOf(".");

      if (idx > -1)
      {
         name = name.substring(0, idx);
      }

      epstopdf(file, new File(file.getParentFile(), name+".pdf"));
   }

   public void epstopdf(File file, File pdfFile)
     throws IOException,InterruptedException
   {
      int exitCode = -1;

      String app = getEpstopdfApp();

      String fileName = file.getAbsolutePath();

      exitCode = execCommandAndWaitFor(new String[]{app, 
         "--outfile="+pdfFile.getAbsolutePath(),
         fileName});

      if (exitCode != 0)
      {
         throw new IOException(getLabelWithValues("error.app_failed",
           app+" \""+fileName+"\"", ""+exitCode));
      }
   }

   public void wmftoeps(File file)
     throws IOException,InterruptedException
   {
      String name = file.getName();

      int idx = name.lastIndexOf(".");

      if (idx > -1)
      {
         name = name.substring(0, idx);
      }

      wmftoeps(file, new File(file.getParentFile(), name+".eps"));
   }

   public void wmftoeps(File file, File pdfFile)
     throws IOException,InterruptedException
   {
      int exitCode = -1;

      String app = getWmftoepsApp();

      String fileName = file.getAbsolutePath();

      exitCode = execCommandAndWaitFor(new String[]{app, 
         "-o", pdfFile.getAbsolutePath(),
         fileName});

      if (exitCode != 0)
      {
         throw new IOException(getLabelWithValues("error.app_failed",
           app+" \""+fileName+"\"", ""+exitCode));
      }
   }

   public String kpsewhich(String name)
     throws IOException,InterruptedException
   {
      return kpsewhich(name, true);
   }

   public String kpsewhich(String name, boolean promptIfAppNotFound)
     throws IOException,InterruptedException
   {
      String value = kpsewhichResults.get(name);

      if (value != null)
      {
         return value;
      }

      int exitCode = -1;

      KpsewhichListener listener = new KpsewhichListener(this);

      String app = getKpsewhichApp(promptIfAppNotFound);

      exitCode = execCommandAndWaitFor(new String[]{app, name},
            listener);

      if (exitCode != 0)
      {
         throw new IOException(getLabelWithValues("error.app_failed",
           app+" "+name, ""+exitCode));
      }

      value = listener.getFile().getAbsolutePath();

      if (value != null && !value.isEmpty())
      {
         kpsewhichResults.put(name, value);
      }

      return value;
   }

   public File getTeXMF()
      throws IOException
   {
      if (texmf != null)
      {
         return texmf;
      }

      // Try to use kpsewhich -var-value=TEXMFHOME to find target
      // directory

      try
      {
         texmf = new File(kpsewhich("-var-value=TEXMFHOME"));
      }
      catch (InterruptedException e)
      {
         error(e);
      }

      return texmf;
   }

   public void fetchJmlrBundle()
      throws IOException,InterruptedException
   {
      try
      {
         disableFunctions();

         File targetDir = getTeXMF();

         // Confirm destination directory

         JFileChooser dirChooser = new JFileChooser(
            targetDir == null ? null : targetDir.getParentFile());
         dirChooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);

         if (targetDir != null)
         {
            dirChooser.setSelectedFile(targetDir);
         }

         if (dirChooser.showDialog(this, getLabel("button.select"))
             != JFileChooser.APPROVE_OPTION)
         {
            return;
         }

         targetDir = dirChooser.getSelectedFile();

         setWaitCursor();

         installJMLRBundle(targetDir);

         String[] cmd = new String[]
         {
            properties.getTeXHash(),
            targetDir.getAbsolutePath()
         };

         String info = cmd[0]+" \""+cmd[1]+"\"";

         setInfo(getLabelWithValue("message.running", info));

         int exitCode = execCommandAndWaitFor(cmd);

         if (exitCode != 0)
         {
            error(getLabelWithValue("error.exec_failed", info));
         }

         setInfo(getLabel("message.done"));
      }
      finally
      {
         resetFunctions();
         setDefaultCursor();
      }
   }

   private void installJMLRBundle(File targetDir)
      throws IOException,InterruptedException
   {
      unpackArchive(
         new URL("http://mirror.ctan.org/install/macros/latex/contrib/jmlr.tds.zip"),
         targetDir);
   }

   // Fetch and unpack zip file
   // Code from
   // http://www.java2s.com/Tutorial/Java/0180__File/UnpackanarchivefromaURL.htm

   public void unpackArchive(URL url, File targetDir)
     throws IOException,InterruptedException
   {
      if (!(targetDir.exists() || targetDir.mkdirs()))
      {
         throw new IOException(getLabelWithValue("error.io.cant_mkdir",
           targetDir.getAbsolutePath()));
      } 

      setInfo(getLabelWithValue("message.ciml.downloading", url.toString()));

      InputStream in = new BufferedInputStream(url.openStream(), 1024);

      Path zip = Files.createTempFile("archive", ".zip");

      Files.copy(in, zip, StandardCopyOption.REPLACE_EXISTING);

      File zipFile = zip.toFile();
      zipFile.deleteOnExit();

      unpackArchive(zipFile, targetDir);
   }

// Unpack entries ending with (case-insensitive) suffix to target
// directory
   public void unpackZipEntries(URL url, File targetDir, String suffix)
     throws IOException,InterruptedException
   {
      if (!(targetDir.exists() || targetDir.mkdirs()))
      {
         throw new IOException(getLabelWithValue("error.io.cant_mkdir",
           targetDir.getAbsolutePath()));
      } 

      suffix = suffix.toLowerCase();

      setInfo(getLabelWithValue("message.ciml.downloading", url.toString()));

      InputStream in = new BufferedInputStream(url.openStream(), 1024);

      Path zip = Files.createTempFile("archive", ".zip");
      zip.toFile().deleteOnExit();

      Files.copy(in, zip, StandardCopyOption.REPLACE_EXISTING);

      ZipFile zipFile = new ZipFile(zip.toFile());

      boolean overWriteAll = false;

      for (Enumeration<? extends ZipEntry> entries = zipFile.entries(); entries.hasMoreElements();)
      {
         ZipEntry entry = entries.nextElement();

         String name = entry.getName();

         if (name.toLowerCase().endsWith(suffix))
         {
            overWriteAll = unpackFile(targetDir, zipFile, entry, overWriteAll);
         }
      }
   }

   public void unpackZipEntry(URL url, File targetDir, String suffix, String fileName)
     throws IOException,InterruptedException
   {
      if (!(targetDir.exists() || targetDir.mkdirs()))
      {
         throw new IOException(getLabelWithValue("error.io.cant_mkdir",
           targetDir.getAbsolutePath()));
      } 

      suffix = suffix.toLowerCase();

      setInfo(getLabelWithValue("message.ciml.downloading", url.toString()));

      InputStream in = new BufferedInputStream(url.openStream(), 1024);

      Path zip = Files.createTempFile("archive", ".zip");
      zip.toFile().deleteOnExit();

      Files.copy(in, zip, StandardCopyOption.REPLACE_EXISTING);

      ZipFile zipFile = new ZipFile(zip.toFile());

      boolean overWriteAll = false;

      for (Enumeration<? extends ZipEntry> entries = zipFile.entries(); entries.hasMoreElements();)
      {
         ZipEntry entry = entries.nextElement();

         String name = entry.getName();

         if (name.toLowerCase().endsWith(suffix))
         {
            overWriteAll = unpackFile(targetDir, zipFile, entry, fileName, overWriteAll);
            break;
         }
      }
   }

   public void unpackArchive(File theFile, File targetDir)
      throws IOException,InterruptedException
   {
      if (!theFile.exists())
      {
         throw new IOException(getLabelWithValue("error.file_not_found",
            theFile.getAbsolutePath()));
      }

      if (!(targetDir.exists() || targetDir.mkdirs()))
      {
         throw new IOException(getLabelWithValue("error.io.cant_mkdir",
           targetDir.getAbsolutePath()));
      }

      boolean overWriteAll = false;

      ZipFile zipFile = new ZipFile(theFile);

      for (Enumeration<? extends ZipEntry> entries = zipFile.entries(); entries.hasMoreElements();)
      {
         ZipEntry entry = entries.nextElement();

         overWriteAll = unpackFile(targetDir, zipFile, entry, overWriteAll);
      }

      zipFile.close();
   }

   public boolean unpackFile(File targetDir, ZipFile zipFile, ZipEntry entry, boolean overWriteAll)
     throws IOException,InterruptedException
   {
      return unpackFile(targetDir, zipFile, entry, entry.getName(), overWriteAll);
   }

   public boolean unpackFile(File targetDir, ZipFile zipFile, ZipEntry entry,
      String fileName, boolean overWriteAll)
     throws IOException,InterruptedException
   {
      File file = new File(targetDir, fileName);

      setInfo(getLabelWithValue("message.unpacking", file.getAbsolutePath()));

      File parent = file.getParentFile();

      if (!(parent.exists() || parent.mkdirs()))
      {
        throw new IOException(getLabelWithValue("error.io.cant_mkdir",
          parent.getAbsolutePath()));
      }

      if (entry.isDirectory())
      {
         if (!(file.exists() || file.mkdirs()))
         {
            throw new IOException(getLabelWithValue("error.io.cant_mkdir",
               file.getAbsolutePath()));
         }
      }
      else
      {
         if (overWriteAll)
         {
            Files.copy(zipFile.getInputStream(entry), file.toPath(),
               StandardCopyOption.REPLACE_EXISTING);
         }
         else if (file.exists())
         {
            // confirm overwrite

            Object thisVal = getLabel("message.overwrite.this");
            Object allVal  = getLabel("message.overwrite.all");
            Object skipVal = getLabel("message.overwrite.skip");

            Object result = JOptionPane.showInputDialog(this, 
               getLabelWithValue("message.overwrite", file.getAbsolutePath()),
               getLabel("message.confirm"),
               JOptionPane.WARNING_MESSAGE,
               null,
               new Object[]
               {
                  thisVal,
                  allVal,
                  skipVal
               },
               thisVal
            );

            if (result == null)
            {
               throw new CancelledException();
            }

            if (result == allVal)
            {
               overWriteAll = true;
            }

            if (result == allVal || result == thisVal)
            {
               Files.copy(zipFile.getInputStream(entry), file.toPath(),
                  StandardCopyOption.REPLACE_EXISTING);
            }
         }
         else
         {
            Files.copy(zipFile.getInputStream(entry), file.toPath());
         }
      }

      return overWriteAll;
   }

   public long getMaxProcessTime()
   {
      if (MAX_PROCESS_TIME > 0L)
      {
         return MAX_PROCESS_TIME;
      }

      return properties.getMaxProcessTime();
   }

   public int execCommandAndWaitFor(String[] cmd)
     throws IOException,InterruptedException
   {
      return execCommandAndWaitFor(cmd, null, null, new DefaultProcessListener(this), true);
   }

   public int execCommandAndWaitFor(String[] cmd, boolean writeInfo)
     throws IOException,InterruptedException
   {
      return execCommandAndWaitFor(cmd, null, null, new DefaultProcessListener(this), writeInfo);
   }

   public int execCommandAndWaitFor(String[] cmd, ProcessListener listener)
     throws IOException,InterruptedException
   {
      return execCommandAndWaitFor(cmd, null, null, listener, true);
   }

   public int execCommandAndWaitFor(String[] cmd, String[] envp, File dir)
     throws IOException,InterruptedException
   {
      return execCommandAndWaitFor(cmd, envp, dir, new DefaultProcessListener(this), true);
   }

   public int execCommandAndWaitFor(String[] cmd, String[] envp, File dir, ProcessListener listener)
     throws IOException,InterruptedException
   {
      return execCommandAndWaitFor(cmd, envp, dir, listener, true);
   }

   // adapted from http://kylecartmell.com/?p=9
   public int execCommandAndWaitFor(String[] cmd, String[] envp, File dir, 
      ProcessListener listener, boolean writeInfo)
     throws IOException,InterruptedException
   {
      java.util.Timer timer = null;
      Process p = null;
      int exitCode = -1;

      InterruptTimerTask interruptor = null;

      try
      {
         timer = new java.util.Timer(true);
         interruptor = new InterruptTimerTask(Thread.currentThread());
         timer.schedule(interruptor, getMaxProcessTime());
         listener.setInterruptor(interruptor);

         p = execCommand(cmd, envp, dir, listener, writeInfo);

         addProcessListener(listener);

         exitCode = p.waitFor();
      }
      catch (InterruptedException e)
      {
         p.destroy();

         throw e;
      }
      finally
      {
         timer.cancel();
         Thread.interrupted();
         removeProcessListener(listener);
         updateAbortItem();
      }

      return exitCode;
   }

   public Process execCommand(String[] cmd)
     throws IOException
   {
      return execCommand(cmd, null, null, new DefaultProcessListener(this), true);
   }

   public Process execCommand(String[] cmd, boolean writeInfo)
     throws IOException
   {
      return execCommand(cmd, null, null, new DefaultProcessListener(this), writeInfo);
   }

   public Process execCommand(String[] cmd, ProcessListener listener)
     throws IOException
   {
      return execCommand(cmd, null, null, listener, true);
   }

   public Process execCommand(String[] cmd, String[] envp, File dir)
     throws IOException
   {
      return execCommand(cmd, envp, dir, new DefaultProcessListener(this), true);
   }

   public Process execCommand(String[] cmd, String[] envp, File dir, ProcessListener listener)
     throws IOException
   {
      return execCommand(cmd, envp, dir, listener, true);
   }

   public Process execCommand(String[] cmd, String[] envp, File dir,
       ProcessListener listener, boolean writeInfo)
     throws IOException
   {
      String params = "";

      for (int i = 1; i < cmd.length; i++)
      {
         params += " \""+cmd[i]+"\"";
      }

      String execName = cmd[0];

      int idx = execName.lastIndexOf(File.separator);

      if (idx > -1)
      {
         execName = execName.substring(idx+1);
      }

      if (debugMode || LOGWRITER != null)
      {
         debug("Running:");

         if (dir != null)
         {
            debug("  in directory "+dir.getAbsolutePath());
         }

         if (envp != null && envp.length > 0)
         {
            debug("  with environment:");

            for (int i = 0; i < envp.length; i++)
            {
               debug("  "+envp[i]);
            }
         }

         debug(cmd[0]+params);
      }

      if (writeInfo)
      {
         setInfo(getLabelWithValue("message.running", execName+params));
      }

      Process p = Runtime.getRuntime().exec(cmd, envp, dir);

      listener.setProcess(p);

      ProcessInputReaderThread inReaderThread = new ProcessInputReaderThread(p, listener);
      listener.setThread(inReaderThread);
      inReaderThread.start();
      inReaderThread = null;

      return p;
   }

//http://www.forward.com.au/javaProgramming/HowToStopAThread.html
   public static synchronized void checkForInterrupt()
     throws InterruptedException
   {
      Thread.yield();

      if (Thread.currentThread().isInterrupted())
      {
         throw new CancelledException();
      }
   }

   public synchronized boolean abort()
   {
      if (JOptionPane.showConfirmDialog(null,
         getLabel("message.confirm.abort"),
         getLabel("message.confirm"),
         JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
      {
         for (Thread thread : currentWorkerThreads)
         {
            debug("Interrupting "+thread+" "+thread.getClass().toString());
            thread.interrupt();
         }

         for (ProcessListener listener : currentProcessListeners)
         {
            listener.terminateProcess();
         }

         abortItem.setEnabled(false);

         currentWorkerThreads.clear();
         currentProcessListeners.clear();

         return true;
      }

      return false;
   }

   public synchronized void addProcessListener(ProcessListener listener)
   {
      if (currentProcessListeners != null)
      {
         currentProcessListeners.add(listener);
      }
   }

   public synchronized void removeProcessListener(ProcessListener listener)
   {
      if (currentProcessListeners != null)
      {
         currentProcessListeners.remove(listener);
      }
   }

   public synchronized void startWorkerThread(Thread thread)
   {
      debug("adding thread to current list");
      currentWorkerThreads.add(thread);
      debug("starting thread");
      thread.start();
      abortItem.setEnabled(true);
   }

   public synchronized void workerThreadFinished(Thread thread)
   {
      if (currentWorkerThreads != null)
      {
         currentWorkerThreads.remove(thread);

         updateAbortItem();
      }
   }

   private synchronized void updateAbortItem()
   {
      if (abortItem != null && currentWorkerThreads != null
       && currentProcessListeners != null)
      {
         abortItem.setEnabled(currentWorkerThreads.size() > 0 ||
            currentProcessListeners.size() > 0);
      }
   }

   private JMenu createMenu(String label)
   {
      return createMenu(null, label);
   }

   private JMenu createMenu(String parentLabel, String label)
   {
      JMenu menu = new JMenu(getLabel(parentLabel, label));
      menu.setMnemonic(getMnemonic(parentLabel, label));

      return menu;
   }

   private JMenuItem createMenuItem(String parentLabel, String label)
   {
      return createMenuItem(parentLabel, label, null, null, this);
   }

   private JMenuItem createMenuItem(String parentLabel, String label, KeyStroke keyStroke)
   {
      return createMenuItem(parentLabel, label, keyStroke, null, this);
   }

   private JMenuItem createMenuItem(String parentLabel, String label, String tooltip)
   {
      return createMenuItem(parentLabel, label, null, tooltip, this);
   }

   private JMenuItem createMenuItem(String parentLabel, String label,
      KeyStroke keyStroke, String tooltip)
   {
      return createMenuItem(parentLabel, label, keyStroke, tooltip, this);
   }

   private JMenuItem createMenuItem(String parentLabel, String label,
      KeyStroke keyStroke, String tooltip, ActionListener listener)
   {
      JMenuItem item = new JMenuItem(getLabel(parentLabel, label));
      item.setMnemonic(getMnemonic(parentLabel, label));
      item.setActionCommand(label);

      if (listener != null)
      {
         item.addActionListener(listener);
      }

      if (keyStroke != null)
      {
         item.setAccelerator(keyStroke);
      }

      if (tooltip != null)
      {
         item.setToolTipText(tooltip);
      }

      return item;
   }

   private MenuButtonItem createMenuButtonItem(String parentLabel, String label,
      KeyStroke keyStroke, String imageName, JMenu menu)
   {
      return createMenuButtonItem(parentLabel, label, keyStroke, imageName, this, menu);
   }

   private MenuButtonItem createMenuButtonItem(String parentLabel, String label,
      KeyStroke keyStroke, String imageName, ActionListener listener, JMenu menu)
   {
      return createMenuButtonItem(parentLabel, label, keyStroke, imageName, listener,
        menu, toolBar);
   }

   public static MenuButtonItem createMenuButtonItem(String parentLabel, String label,
      KeyStroke keyStroke, String imageName, ActionListener listener, JMenu menu,
      JComponent toolBarComp)
   {
      String tooltip = dictionary.getProperty(parentLabel+"."+label+".tooltip");
      String altText = dictionary.getProperty(parentLabel+"."+label+".altText");

      String imgLocation;

      if (imageName.endsWith(".png"))
      {
         imgLocation = "icons/"+imageName;
      }
      else
      {
         imgLocation = "/toolbarButtonGraphics/"+imageName+".gif";
      }

      URL imageURL = MakeJmlrBookGUI.class.getResource(imgLocation);

      MenuButtonItem item = new MenuButtonItem(getLabel(parentLabel, label),
        getMnemonic(parentLabel, label), imageURL, label,
        keyStroke, listener, tooltip, altText);

      toolBarComp.add(item.getButton());
      menu.add(item.getItem());

      return item;
   }

   public static String getTeXPath(File baseDir, File file)
   {
      return getTeXPath(baseDir.toPath(), file.toPath());
   }

   public static String getTeXPath(Path basePath, Path filePath)
   {
      Path relPath = basePath.relativize(filePath);

      String path = null;

      for (Iterator<Path> it = relPath.iterator(); it.hasNext();)
      {
         Path p = it.next();

         if (path == null)
         {
            path = p.toString();
         }
         else
         {
            path += "/" + p.toString();
         }
      }

      return path;
   }

   public static String getLabelWithAlt(String label, String alt)
   {
      if (dictionary == null) return alt;

      String prop = dictionary.getProperty(label);

      if (prop == null)
      {
         return alt;
      }

      return prop;
   }

   public static String getLabel(String label)
   {
      return getLabel(null, label);
   }

   public static String getLabel(String parent, String label)
   {
      if (parent != null)
      {
         label = parent+"."+label;
      }

      String prop = dictionary.getProperty(label);

      if (prop == null)
      {
         System.err.println("No such dictionary property '"+label+"'");
         return "?"+label+"?";
      }

      return prop;
   }

   public static char getMnemonic(String label)
   {
      return getMnemonic(null, label);
   }

   public static char getMnemonic(String parent, String label)
   {
      String prop = getLabel(parent, label+".mnemonic");

      if (prop.equals(""))
      {
         System.err.println("Empty dictionary property '"+prop+"'");
         return label.charAt(0);
      }

      return prop.charAt(0);
   }

   public static int getMnemonicInt(String label)
   {
      return getMnemonicInt(null, label);
   }

   public static int getMnemonicInt(String parent, String label)
   {
      String prop = getLabel(parent, label+".mnemonic");

      if (prop.equals(""))
      {
         System.err.println("Empty dictionary property '"+prop+"'");
         return label.codePointAt(0);
      }

      return prop.codePointAt(0);
   }

   public static String getLabelWithValue(String label, String value)
   {
      return getLabelWithValues(label, new String[]{value});
   }

   public static String getLabelWithValue(String label, int value)
   {
      return getLabelWithValue(label, ""+value);
   }

   public static String getLabelWithValues(String label, String value1,
      String value2)
   {
      return getLabelWithValues(label, new String[]{value1, value2});
   }

   public static String getLabelWithValues(String label, String[] values)
   {
      String prop = getLabel(label);

      for (int i = 0; i < values.length; i++)
      {
         String tag = "$"+(i+1);
         int idx = prop.indexOf(tag);

         while (idx > -1)
         {
            String pre = "";

            if (idx > 0)
            {
               pre = prop.substring(0, idx);
            }

            String post = prop.substring(idx+tag.length());

            prop = pre + values[i] + post;

            idx = prop.indexOf(tag);
         }
      }

      return prop;
   }

   private void initHelp()
   {
      if (mainHelpBroker == null)
      {
         HelpSet mainHelpSet = null;

         String resource = "makejmlrbookgui";

         String helpsetLocation = "/resources/helpsets/"+resource;

         String lang = Locale.getDefault().getLanguage();

         try
         {
            URL hsURL = getClass().getResource(helpsetLocation+"-"+lang+"/"+resource+".hs");

            if (hsURL == null && !lang.equals("en"))
            {
               hsURL = getClass().getResource(helpsetLocation+"-en/"+resource+".hs");
            }

            mainHelpSet = new HelpSet(null, hsURL);
         }
         catch (Exception e)
         {
            JMLRResources.error(null, "/resources/helpsets/"+resource+".hs\n"+
              getLabel("error.io.helpset")+":\n" +e.getMessage());
         }

         if (mainHelpSet != null)
         {
            mainHelpBroker = mainHelpSet.createHelpBroker();
         }

         if (mainHelpBroker != null)
         {
            csh = new CSH.DisplayHelpFromSource(mainHelpBroker);
         }
      }
   }

   public void enableHelpOnButton(JComponent comp, String id)
   {
      if (mainHelpBroker != null)
      {
         try
         {
            mainHelpBroker.enableHelpOnButton(comp, id,
               mainHelpBroker.getHelpSet());
         }
         catch (BadIDException e)
         {
            JMLRResources.error(null, e);
         }
      }
      else
      {
         System.err.println("Can't enable help on button (id="+id
           +"): null help broker");
      }
   }

   private static void loadDictionary()
      throws IOException
   {
      Locale locale = Locale.getDefault();

      String lang = locale.getLanguage();

      String resource = "makejmlrbook";

      InputStream in =
         MakeJmlrBookGUI.class.getResourceAsStream("/resources/dictionaries/"+resource+"-"+lang+".prop");

      if (in == null && !lang.equals("en"))
      {         in = MakeJmlrBookGUI.class.getResourceAsStream("/resources/dictionaries/"+resource+"-en.prop");
      }

      if (in == null)
      {
         throw new FileNotFoundException
            ("Can't find dictionary resource file /resources/dictionaries/"+resource+"-en");
      }

      BufferedReader reader = new BufferedReader(new InputStreamReader(in));
      dictionary = new Properties();
      dictionary.load(reader);

      reader.close();

      in.close();
   }

   public static void debug(String message)
   {
      if (debugMode)
      {
         System.out.println(message);
      }

      if (LOGWRITER != null)
      {
         LOGWRITER.println(message);
      }
   }

   public static void debug(Throwable e)
   {
      if (debugMode)
      {
         e.printStackTrace();
      }

      if (LOGWRITER != null)
      {
         LOGWRITER.println(e);
      }
   }

   public static void printAndLogMessage(String message)
   {
      System.out.println(message);

      if (LOGWRITER != null)
      {
         LOGWRITER.println(message);
      }
   }

   public static boolean inDebugMode()
   {
      return debugMode;
   }

   private static void createAndShow()
   {
      MakeJmlrBookGUI app = new MakeJmlrBookGUI();
      app.setVisible(true);
   }

   public static void main(String[] args)
   {
      try
      {
         loadDictionary();
      }
      catch (IOException e)
      {
         JMLRResources.error(null, e);
         System.exit(1);
      }

      try
      {
         for (int i = 0; i < args.length; i++)
         {
            if (args[i].equals("-debug"))
            {
               debugMode = true;
            }
            else if (args[i].equals("-nodebug"))
            {
               debugMode = false;
            }
            else if (args[i].equals("-timeout"))
            {
               i++;

               if (i == args.length)
               {
                  throw new IllegalArgumentException("-timeout requires numerical argument.");
               }
   
               try
               {
                  MAX_PROCESS_TIME = Long.parseLong(args[i]);
               }
               catch (NumberFormatException e)
               {
                  throw new IllegalArgumentException(
                     "-timeout requires numerical argument. Found: '"+args[i]+"'",
                     e);
               }
            }
            else if (args[i].equals("-version") || args[i].equals("--version"))
            {
               System.out.println(appName+" version "+appVersion);
               System.exit(0);
            }
            else if (args[i].equals("-help") || args[i].equals("--help"))
            {
               System.out.println(appName+" version "+appVersion);
               System.out.println("");
               System.out.println("Syntax:");
               System.out.println("");
               System.out.println("makejmlrbook-gui [options]");
               System.out.println("");
               System.out.println("Options:");
               System.out.println("");
               System.out.println("-debug\t\t\tSwitch on debug mode");
               System.out.println("-nodebug\t\tSwitch off debug mode (default)");
               System.out.println("-timeout <n>\t\tSet the maximum length of time");
               System.out.println("\t\t\tfor processes for this session,");
               System.out.println("\t\t\twhere <n> is an integer representing");
               System.out.println("\t\t\tthe number of milliseconds");
               System.out.println("\t\t\tuntil timeout.");
               System.out.println("-version\t\tDisplay version number and exit");
               System.exit(0);
            }
            else
            {
               throw new IllegalArgumentException
                  ("Unknown option '"+args[i]+"'.");
            }
         }
      }
      catch (Exception e)
      {
         String message = appName+": "+e.getMessage()+" Use --help for help.";
         JMLRResources.error(null, message);
         System.err.println(message);
         System.exit(1);
      }

      SwingUtilities.invokeLater(new Runnable()
        {
           public void run()
           {
              createAndShow();
           }
        });
   }

   public static final String appVersion = "0.9.3b";
   public static final String appName = "makejmlrbookgui";
   public static final String appDate = "2015-11-15";

   private static boolean debugMode = false;
   private static Properties dictionary;

   private JToolBar toolBar;

   private JTabbedPane tabbedPane;

   private MessageLabel messageArea;

   private JMLRBookPanel mainPanel;

   private FileViewer fileViewer;

   private BookRoot bookRoot;

   private JMLRBookProperties properties;

   private HelpBroker mainHelpBroker;
   private CSH.DisplayHelpFromSource csh;

   private JFileChooser fileChooser, pdfFileChooser, htmlFileChooser, 
     dirChooser, bibFileChooser, zipFileChooser;

   private TeXFileFilter texFileFilter;

   private ImageFileFilter imageFileFilter;

   private BookContentsFilter bookContentsFilter;

   private HtmlFileFilter htmlFileFilter;

   private BibFileFilter bibFileFilter;

   private ZipFileFilter zipFileFilter;

   private JMenuItem clearItem, clearRecentItem, clearRecentCiMLItem, pdfxItem,
     charCheckItem, createBibItem, createZipItem;

   private MenuButtonItem refreshItem, openItem, newItem,
      buildMainItem, buildPrintItem, removeAuxItem,
      abortItem, exportItem;

   private JMenu buildM, recentM, buildPapersM, toolsM;

   private ActionListener recentFilesListener;

   private CharCheckSelector charCheckSelector;

   private NewBookDialog newBookDialog;

   private PropertiesDialog propertiesDialog;

   private JMLRAppSelector appSelector;

   private InformationDialog informationDialog;

   private File texmf = null;

   public static long MAX_PROCESS_TIME=0L;

   public static final int MAX_RERUN = 6;

   private static PrintWriter LOGWRITER=null;

   private Vector<Thread> currentWorkerThreads;

   private Vector<ProcessListener> currentProcessListeners;

   private HashMap<String,String> kpsewhichResults;

   private L2HStringConverter l2hStringConverter;
}
