/*
    Copyright (C) 2013 Nicola L.C. Talbot
    www.dickimaw-books.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/
package com.dickimawbooks.makejmlrbookgui.newbook;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.net.MalformedURLException;
import java.nio.file.Path;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.regex.*;
import java.util.Calendar;
import java.text.Normalizer;
import java.nio.charset.Charset;
import java.nio.charset.IllegalCharsetNameException;

import javax.swing.*;
import javax.swing.table.TableModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import com.dickimawbooks.texparserlib.*;
import com.dickimawbooks.texparserlib.latex.*;
import com.dickimawbooks.texparserlib.latex.jmlr.*;
import com.dickimawbooks.texparserlib.latex.inputenc.InputEncSty;
import com.dickimawbooks.texparserlib.latex2latex.*;
import com.dickimawbooks.texparserlib.bib.*;
import com.dickimawbooks.texparserlib.html.L2HStringConverter;

import com.dickimawbooks.makejmlrbookgui.*;

public class NewBookDialog extends JDialog
  implements ActionListener
{
   public NewBookDialog(MakeJmlrBookGUI application)
   {
      super(application, application.getLabel("newbook.title"), true);

      app = application;

      defaultYear = Calendar.getInstance().get(Calendar.YEAR);

      cardLayout = new CardLayout();

      mainPanel = new JPanel(cardLayout);

      introPanel = new IntroPanel();
      mainPanel.add(introPanel);

      titleInfoPanel = new TitleInfoPanel(this);
      mainPanel.add(titleInfoPanel);

      frontMatterPanel = new FrontMatterPanel(this);
      mainPanel.add(frontMatterPanel);

      bookPartPanel = new BookPartPanel(this);
      mainPanel.add(bookPartPanel);

      articlesPanel = new ArticlesPanel(bookPartPanel, this);
      mainPanel.add(new JScrollPane(articlesPanel));

      bookDirPanel = new BookDirectoryPanel(this);
      mainPanel.add(bookDirPanel);

      processPanel = new ProcessPanel(this);
      mainPanel.add(processPanel);

      getContentPane().add(mainPanel, BorderLayout.CENTER);

      backButton = createActionButton("back",
         KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK));

      nextButton = createActionButton("next",
         KeyStroke.getKeyStroke(KeyEvent.VK_B, InputEvent.CTRL_MASK));

      Box buttonPanel = Box.createHorizontalBox();

      getContentPane().add(buttonPanel, BorderLayout.SOUTH);

      buttonPanel.add(backButton);
      buttonPanel.add(Box.createHorizontalGlue());

      JButton helpButton = JMLRResources.createActionButton(
        "button", "help", null, KeyStroke.getKeyStroke(KeyEvent.VK_F1, 0));

      app.enableHelpOnButton( helpButton, "newbook");

      buttonPanel.add(helpButton);

      buttonPanel.add(Box.createHorizontalGlue());
      buttonPanel.add(nextButton);

      pack();
   }

   private JButton createActionButton(String tag)
   {
      return createActionButton(tag, null);
   }

   private JButton createActionButton(String tag,
     KeyStroke keyStroke)
   {
      return JMLRResources.createActionButton(
        app.getLabel("newbook."+tag),
        app.getMnemonic("newbook."+tag),
        tag, this, keyStroke,
        app.getLabelWithAlt("newbook."+tag+".tooltip", null));
   }

   public void display()
   {
      setLocationRelativeTo(app);
      cardLayout.first(mainPanel);
      currentComponentIndex = 0;
      backButton.setVisible(false);
      nextButton.setVisible(true);

      setVisible(true);
   }

   public NewBookPanel getCurrentPanel()
   {
      NewBookPanel currentPanel = null;

      Component component = mainPanel.getComponent(currentComponentIndex);

      if (component instanceof JScrollPane)
      {
         currentPanel = (NewBookPanel)
                 ((JScrollPane)component).getViewport().getView();
      }
      else
      {
         currentPanel = (NewBookPanel)component;
      }

      return currentPanel;
   }

   public void actionPerformed(ActionEvent evt)
   {
      String action = evt.getActionCommand();

      if (action == null) return;

      if (action.equals("next"))
      {
         NewBookPanel currentPanel = getCurrentPanel();

         if (currentPanel.canMoveForwards())
         {
            cardLayout.next(mainPanel);
            backButton.setVisible(true);
            currentComponentIndex++;

            currentPanel = getCurrentPanel();

            currentPanel.updatePanel();
            nextButton.setVisible(currentPanel.hasNextButton());
            backButton.setVisible(currentPanel.hasBackButton());
         }
      }
      else if (action.equals("back"))
      {
         NewBookPanel currentPanel = getCurrentPanel();

         if (currentComponentIndex > 0 && currentPanel.hasBackButton())
         {
            cardLayout.previous(mainPanel);
            currentComponentIndex--;

            currentPanel = getCurrentPanel();

            backButton.setVisible(currentPanel.hasBackButton());
            nextButton.setVisible(currentPanel.hasNextButton());
         }
      }
   }

   public boolean isCiML()
   {
      return titleInfoPanel.isCiML();
   }

   public int getVolume()
   {
      return isCiML() ? titleInfoPanel.getCiMLVolume() :
         titleInfoPanel.getJMLRVolume();
   }

   public int getWCPVolume()
   {
      return isCiML() ? titleInfoPanel.getCiMLWCPVolume() :
         titleInfoPanel.getJMLRVolume();
   }

   public int getYear()
   {
      return isCiML() ? defaultYear : titleInfoPanel.getJMLRYear();
   }

   public String getMonth()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRMonth();
   }

   public String getDay()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRDay();
   }

   public String getJMLRLocation()
   {
      return isCiML() ? null : titleInfoPanel.getJMLRLocation();
   }

   public Vector<Contributor> getEditors()
   {
      return isCiML() ? titleInfoPanel.getCiMLEditors() :
         titleInfoPanel.getJMLREditors();
   }

   public String getBookTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLTitle() :
         titleInfoPanel.getJMLRTitle();
   }

   public String getBookShortTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLShortTitle() :
         titleInfoPanel.getJMLRShortTitle();
   }

   public String getBookSubTitle()
   {
      return isCiML() ? titleInfoPanel.getCiMLSubTitle() : "";
   }

   public String getProductionEditor()
   {
      return isCiML() ? titleInfoPanel.getCiMLProductionEditor() : "";
   }

   public File getBookDirectory()
   {
      return bookDirPanel.getBookDirectory();
   }

   public String getBookFonts()
   {
      return bookDirPanel.getBookFonts();
   }

   public Vector<PartListPanel> getPartLists()
   {
      return articlesPanel.getPartLists();
   }

   public ImportPapersPanel getMainImportPapers()
   {
      return articlesPanel.getMainImportPapers();
   }

   public ImportPapersPanel getAppendixImportPapers()
   {
      return articlesPanel.getAppendixImportPapers();
   }

   public boolean hasPreface()
   {
      return frontMatterPanel.hasPreface();
   }

   public boolean hasForeword()
   {
      return frontMatterPanel.hasForeword();
   }

   public boolean hasAppendices()
   {
      return bookPartPanel.hasAppendices();
   }

   public int getPartConfig()
   {
      return bookPartPanel.getPartConfig();
   }

   public File getPrefaceFile()
   {
      return frontMatterPanel.getPrefaceFile();
   }

   public void setPrefaceFile(File file)
   {
      frontMatterPanel.setPrefaceFile(file);
   }

   public File getForewordFile()
   {
      return frontMatterPanel.getForewordFile();
   }

   public void addCimlImportFile(File base, String section, BibEntry entry,
     TeXParser parser)
   {
      articlesPanel.addCimlImportFile(base, section, entry, parser);
   }

   public JFileChooser getFileChooser()
   {
      return app.getFileChooser();
   }

   public JFileChooser getDirChooser()
   {
      return app.getDirChooser();
   }

   public ImageFileFilter getImageFileFilter()
   {
      return app.getImageFileFilter();
   }

   public JFileChooser getBibFileChooser()
   {
      return app.getBibFileChooser();
   }

   public MakeJmlrBookGUI getApplication()
   {
      return app;
   }

   public File getPackageLocation(String name)
   {
      try
      {
         return new File(app.kpsewhich(name+".sty"));
      }
      catch (IOException e)
      {
      }
      catch (InterruptedException e)
      {
      }

      return null;
   }

   public void verboseCopyFile(File srcFile, File destFile)
    throws IOException
   {
      app.verboseCopyFile(srcFile, destFile);
   }

   public String kpsewhich(String name)
     throws IOException
   {
      try
      {
         return app.kpsewhich(name);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void epstopdf(File epsFile, File pdfFile)
     throws IOException
   {
      try
      {
         app.epstopdf(epsFile, pdfFile);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void wmftoeps(File wmfFile, File epsFile)
     throws IOException
   {
      try
      {
         app.wmftoeps(wmfFile, epsFile);
      }
      catch (InterruptedException e)
      {
         throw new IOException(MakeJmlrBookGUI.getLabel("error.interrupted"),
           e);
      }
   }

   public void viewFile(File file)
     throws MalformedURLException
   {
      app.viewFile(file);
   }

   public void setAppInfo(String msg)
   {
      app.setInfo(msg);
   }

   public void finish(File file)
   {
      setVisible(false);

      app.setInfo(MakeJmlrBookGUI.getLabel("message.done"));

      if (file != null && file.exists())
      {
         try
         {
            app.load(file);
         }
         catch (IOException e)
         {
            app.error(e);
         }
      }
   }

   public void addPart(String title)
   {
      if (!bookPartPanel.hasParts())
      {
         articlesPanel.removeEmptyDefaultParts();
         bookPartPanel.setParts(true);
      }

      articlesPanel.addPart().setPartTitle(title);
      articlesPanel.revalidate();
   }

   public void clearImportData()
   {
      articlesPanel.clearImportData();
   }

   private MakeJmlrBookGUI app;

   private CardLayout cardLayout;

   private JComponent mainPanel;

   private IntroPanel introPanel;
   private TitleInfoPanel titleInfoPanel;
   private FrontMatterPanel frontMatterPanel;
   private BookPartPanel bookPartPanel;
   private ArticlesPanel articlesPanel;
   private BookDirectoryPanel bookDirPanel;
   private ProcessPanel processPanel;

   private int currentComponentIndex;

   private int defaultYear=0;

   private JButton nextButton, backButton;
}
